# -*- coding: utf-8 -*-

__doc__ = """This module implements the wait.
"""

from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


CONDITION_DIC = {
    "presence": EC.presence_of_element_located,
    "clickable": EC.element_to_be_clickable,
    "selected": EC.element_to_be_selected,
    "presence_all": EC.presence_of_all_elements_located,
    "visibility": EC.visibility_of_element_located,
    "visibility_all": EC.visibility_of_all_elements_located,
    "alert_presence": EC.alert_is_present,
}


def element_wait(
        element,
        time: int,
        by_strat: str,
        val: str,
        condition_type: str
) -> webdriver.firefox.webelement.FirefoxWebElement:
    """Tell a selenium webelement to wait for `time` seconds
    until the val in element is located.

    Parameters
    ----------
    element : selenium.webdriver.webelement
        The element which is searched for val given the
        condition type and the by strategy.
    time : int
        The amount of time in seconds which should be waited.
    by_strat : str
        The by-strategy chosen.
    val : str
        The value which is searched in element given the by strategy.
    condition_type : str
        The type of condition to be applied.
        If no valid condition is given, the element waits until the value in
        element is present. Valid conditions are:

        * "presence": Wait until value in element is present.
        * "clickable": Wait until value in element is clickable.
        * "selected": Wait until value in element is selected.
        * "presence_all": Wait until all values in element are present.
        * "visibility": Wait until value in element is visible.
        * "visibility_all": Wait until all values in element are visible.
        * "alert_presence": Wait until an alert is present.

    Returns
    -------
    list
        A list containing the element(s) which was/were found,
        if no element was found, either None or an empty list is returned,
        depending on the condition chosen.

    Raises
    ------
    selenium.common.exceptions.TimeoutException
        If no element is found in the given `time`,
        then a TimeoutException is raised.

    """
    condition = CONDITION_DIC.get(
        condition_type, EC.presence_of_element_located)
    try:
        element = WebDriverWait(element, time).until(
            condition((by_strat, val)))
    except TimeoutException as timeout:
        msg = f"The value {val} which should be searched in {element} " \
              f"could not be located in {time} " \
              f"seconds given the by strategy {by_strat} " \
              f"and the condition {condition}."
        raise TimeoutException(msg) from timeout
    return element



# -*- coding: utf-8 -*-

__doc__ = """This module implements some scrapers.
"""

from pathlib import Path

__version__ = "0.0.6"

# define the log dir which is created in the same folder this projcet is ran
LOG_DIR = Path("./log")
if not LOG_DIR.exists():
    LOG_DIR.mkdir(parents=True, exist_ok=True)



# -*- coding: utf-8 -*-

__doc__ = """This module implements the webdriver.

The selenium based websdriver is useful for navigating through
webpages.
"""

import os
from typing import Optional, Union

from selenium import webdriver
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        ElementNotInteractableException)
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options

from ._base import Scraper
from .wait import element_wait


class Webdriver(Scraper):
    """The `Webdriver` class.
    """

    def __init__(self, url: str, exe_path: Union[os.PathLike, str], headless: bool = True) -> None:
        """Init the class.

        Parameters
        ----------
        url : str
            The url which should be loaded.
        exe_path : Union[PathLike, str]
            The path to the executable browser.
        headless : bool, optional
            Determine if the browser should be executed
            without opening a window, by default True.

        """
        super().__init__()
        # define the path to the driver
        self._path = exe_path
        # set some options using the built-in Options class
        options = Options()
        options.headless = headless
        # define the engine, i.e. the browser to be used
        self._engine = webdriver.Firefox(
            options=options, executable_path=self._path)
        self._url = url
        # define a strategy dictionary
        self._strategy_dic = {
            "id": By.ID,
            "xpath": By.XPATH,
            "link_text": By.LINK_TEXT,
            "partial_link_test": By.PARTIAL_LINK_TEXT,
            "name": By.NAME,
            "tag_name": By.TAG_NAME,
            "class_name": By.CLASS_NAME,
            "css_selector": By.CSS_SELECTOR,
        }
        # define a timeout
        self._timeout = 10

    def __enter__(self) -> None:
        self.get()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context manager and ensure that the connection is closed.

        Parameters
        ----------
        exc_type : [type]
            The  execution type.
        exc_val : [type]
            The execution value.
        exc_tb : [type]
            The execution traceback.

        """
        self._engine.quit()

    @property
    def engine(self) -> webdriver.Firefox:
        """Get the current engine.

        Returns
        -------
        webdriver
            The current engine.

        """
        return self._engine

    def click_hyperlink(self, hyperlink) -> None:
        """Click a hyperlink object.

        Parameters
        ----------
        hyperlink : selenium.webdriver.Webelement
            A selenium webelement representing a hyperlink.

        References
        ----------
        [1] https://sqa.stackexchange.com/questions/13792/how-to-proceed-after-clicking-a-link-to-new-page-in-selenium-in-python

        """
        try:
            hyperlink.click()
        except ElementClickInterceptedException:
            self._engine.execute_script("arguments[0].click();", hyperlink)
        except ElementNotInteractableException:
            self._engine.execute_script(
                "return arguments[0].scrollIntoView(true);", hyperlink)
            self._engine.execute_script("arguments[0].click();", hyperlink)
        finally:
            window_after = next(
                iter(self._engine.window_handles), self._engine.window_handles[0])
            self._engine.switch_to.window(window_after)

    def parse(
            self,
            by_strat: Optional[str],
            **kwargs: dict
    ) -> list:
        """Find all <object></object> html elements in the given
        html page or in the given webelement.

        Parameters
        ----------
        by_strat : str or None
            Determine a by-strategy by which the given element
            should be searched for value.
            Can be one of the following:
            * "id"
            * "xpath"
            * "link_text"
            * "partial_link_test"
            * "name"
            * "tag_name"
            * "class_name"
            * "css_selector"

        Other Parameters
        ----------------
        element : selenium.webdriver.webelement
            The webelement which should be searched.
        val : str or None
            The value of the table element corresponding to the type
            of strategy chosen, see [1]. If `val` is None, then the html
            is returned.
        condition_type : str
            See the docs for `element_wait()`.

        Returns
        -------
        list
            A list containing the selenium.webdriver.webelements.
            If the index, i.e. `idx` is set to a single value,
            i.e. an integer, then the webdriver element is directly returned.

        Notes
        -----
        - This function extracts all content in the table:

        .. highlight:: html
        .. code-block:: html

            <object>
                content
            </object>

        - If no by strategy is given, the default is set to CSS_SELECTOR.
        - If a specific element is chosen, the val parameter has to be adapted
          to the by strategy, which is set to CSS_SELECTOR if set to None,
          thus the val parameter has to be as CSS_SELECTOR.
        - If no kwargs are given, the whole html element is searched for <object></object>.

        References
        ----------
        [1] https://selenium-python.readthedocs.io/locating-elements.html

        """
        by_strat = self._strategy_dic.get(by_strat, By.CSS_SELECTOR)
        element = kwargs.get("element", self._engine)
        val = kwargs.get("val", "html")
        condition_type = kwargs.get("condition_type", "presence_all")
        # get the object
        obj = element_wait(
            element=element,
            time=self._timeout,
            by_strat=by_strat,
            val=val,
            condition_type=condition_type
        )
        return obj

    def get(self) -> None:
        """Load the given url.

        Notes
        -----
        This method has to be called before making any other operations
        on the webpage, otherwise the url will not be
        loaded into to driver object.

        """
        self._engine.get(self._url)



# -*- coding: utf-8 -*-

__doc__ = """This module implements the webscraper.
"""

import http.client
import logging
import os
from concurrent.futures import ThreadPoolExecutor  # ,ProcessPoolExecutor
from typing import List, Optional, Union

import requests
from bs4 import BeautifulSoup, SoupStrainer
from bs4.dammit import EncodingDetector
from fake_useragent import UserAgent

from . import LOG_DIR
from ._base import DATA_OBJECT, Scraper

# set up the logging configuration
http.client.HTTPConnection.debuglevel = 0
LOG_FILE = LOG_DIR / f"{__name__.split('.')[-1]}.log"

REQUESTS_LOG = logging.getLogger("requests.packages.urllib3")
REQUESTS_LOG.setLevel(logging.DEBUG)
REQUESTS_LOG.propagate = True

FILE_HANDLER = logging.FileHandler(LOG_FILE, mode="w", encoding="utf-8")
FILEFORMAT = logging.Formatter(
    "%(asctime)s:[%(threadName)-12.12s]:%(levelname)s:%(name)s:%(message)s",
    datefmt="%Y-%m-%d %H:%M:%S"
)
FILE_HANDLER.setFormatter(FILEFORMAT)

STREAM_HANDLER = logging.StreamHandler()
STREAM_HANDLER.setLevel(logging.WARNING)
STREAMFORMAT = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")
STREAM_HANDLER.setFormatter(STREAMFORMAT)

REQUESTS_LOG.addHandler(FILE_HANDLER)
REQUESTS_LOG.addHandler(STREAM_HANDLER)

logging.basicConfig(handlers=[FILE_HANDLER, STREAM_HANDLER])
logging.getLogger().setLevel(logging.WARNING)

STATUS_CODES = {
    100: "Informational Responses",
    200: "Success",
    300: "Redirection",
    400: "Client Errors",
    500: "Server Errors",
}

RESPONSE_OBJECT = Union[None, requests.Response,
                        List[requests.Response], requests.exceptions.RequestException]


def callback(res: requests.Response, *args, **kwargs) -> requests.Response:
    """Callback function.

    Parameters
    ----------
    res : requests.Response
        A response object.

    Returns
    -------
    requests.Response
        A request.Response object.

    """
    # indicate that the callback funtion was called
    res.hook_called = True
    if args:
        raise AssertionError(f"Have a look at what is in {args}")
    msg = f"\n----- REPORT START -----\n" \
          f"URL: {res.url}\n" \
          f"Time: {res.elapsed.total_seconds():.3f}s\n" \
          f"Encoding: {res.encoding}\n" \
          f"Reason: {res.reason}\n" \
          f"Status Code: {res.status_code}\n" \
          f"Certificate: {kwargs.get('cert', None)}\n" \
          f"----- REPORT END -----\n"
    REQUESTS_LOG.debug(msg)
    return res


class Webscraper(Scraper):
    """The Webscraper class.
    """

    def __init__(
        self,
        parser: str,
        verbose: bool = False,
        get_params: Optional[dict] = None,
    ) -> None:
        """Init the class.

        Parameters
        ----------
        parser : str
            The parser to be used. Can either be:

            * "html.parser"
            * "lxml"

        verbose : bool
            Determine whether the output should be written to the log file,
            by default False.
        get_params : Optional[dict]
            A dictionary containing parameters for the GET request,
            by default None. See [1].

        References
        ----------
        [1] https://2.python-requests.org/en/master/

        """
        super().__init__()
        self._parser = parser
        self._verbose = verbose
        if get_params:
            self._get_params = get_params
        else:
            self._get_params = {}

        self._max_threads = os.cpu_count()*2 - 4
        self._max_processes = os.cpu_count() - 2

        self._user_agent = UserAgent()
        self._headers = {"User-Agent": self._user_agent.random}
        self._sess = requests.Session()
        if self._verbose:
            self._sess.hooks["response"].append(callback)
        self._timeout = 15

        # initialize the response and data attributes
        self._res = None
        self._data = None

        # define a dictionary which contains values to check which type of
        # http request has been made
        self._http_request = {
            "DELETE": False,
            "GET": False,
            "PATCH": False,
            "POST": False,
            "PUT": False,
            "OPTIONS": False,
        }

    def __str__(self) -> str:
        if isinstance(self._url, list):
            msg = ""
            for url in self._url:
                msg += url + "\n"
        elif isinstance(self._url, str):
            msg = self._url + "\n"
        else:
            msg = "No url given."
        return msg

    @property
    def res(self) -> RESPONSE_OBJECT:
        """The response object.

        Returns
        -------
        RESPONSE_OBJECT
            None if not yet set, the/all response object(s),
            or the error thrown.

        """
        return self._res

    @property
    def data(self) -> DATA_OBJECT:
        """The data object.

        Returns
        -------
        DATA_OBJECT
            The data object.

        """
        return self._data

    @data.setter
    def data(self, val: DATA_OBJECT) -> None:
        """Set a new data object.

        Parameters
        ----------
        val : DATA_OBJECT
            The new value for data.

        """
        self._data = val

    def get(
        self,
        url: Union[str, List[str]],
    ) -> None:
        """Make a GET request to a single or multiple urls.

        Parameters
        ----------
        url : Union[str, List[str]]
            The url or list of urls to be loaded.

        Raises
        ------
        AssertionError
            If `url` is neither of type `str` nor of type `list`.

        """
        # save the url to the data class
        self._url = url
        if isinstance(self._url, str):
            self._url = self._url.strip()
            res = self._load_url(self._url)
        elif isinstance(self._url, list):
            self._url = [ur.strip() for ur in self._url]
            res = self._load_urls(self._url)
        else:
            raise AssertionError(
                f"Parameter url is neither of type {str} nor {list}, it is of type {type(url)}.")
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["GET"] = True

    def _load_url(
        self,
        url: str,
    ) -> Union[requests.Response, requests.exceptions.RequestException]:
        """Load a single url.

        Parameters
        ----------
        url : str
            The url to be loaded.

        Returns
        -------
        Union[requests.Response, requests.exceptions.RequestException]
            The corresponding response object. If the request fails
            for some reason, the error itself is returned.

        """
        with self._sess as sess:
            try:
                res = sess.get(
                    url,
                    headers=self._headers,
                    timeout=self._timeout,
                    **self._get_params,
                )
                if not res.ok:  # check if no bad response is returned
                    REQUESTS_LOG.warning(
                        f"Response for {url} failed!\nResponse status code: {res.status_code}.")
                if self._verbose:
                    REQUESTS_LOG.debug("Total Time: %3f s",
                                       res.elapsed.total_seconds())
            except requests.exceptions.RequestException as e:
                REQUESTS_LOG.warning(
                    f"Sending a GET request to {url} has failed!\nThe exception thrown is {e}")
                res = e
        return res

    def _load_urls(
        self,
        urls: List[str],
    ) -> List[requests.Response]:
        """Load a list of urls to response objects.

        Parameters
        ----------
        urls : list
            A list of urls to be loaded with the session objects.

        Returns
        -------
        List[requests.Response]
            A list of requests.Response objects corresponding to the urls given.

        Notes
        -----
        This function uses multithreading since loading multiple URLs is an I/O
        bound task. For this, a computer and system dependent maximum number
        of threads have to be given.

        """
        with ThreadPoolExecutor(max_workers=self._max_threads) as executor:
            responses = list(executor.map(self._load_url, urls, chunksize=8))
            # wait until all threads are finished
            executor.shutdown(wait=True)
        # filter out Response >=[400] and exceptions
        _res = []
        _urls = []
        for i, res in enumerate(responses):
            # filter out errors
            if isinstance(res, requests.exceptions.RequestException):
                continue
            else:
                if res.ok:
                    _res.append(res)
                    _urls.append(urls[i])
                else:  # filter out bad responses
                    continue
        self._url = _urls
        return _res

    def put(
        self,
        url: str
    ) -> None:
        """Make a PUT request to a single url.

        Parameters
        ----------
        url : str
            The url to which a PUT request should be made.

        """
        self._url = url
        with self._sess as sess:
            res = sess.put(url)
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["PUT"] = True

    def delete(
        self,
        url: str
    ) -> None:
        """Make a DELETE request to a single url.

        Parameters
        ----------
        url : str
            The url to which a DELETE request should be made.

        """
        self._url = url
        with self._sess as sess:
            res = sess.delete(url)
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["DELETE"] = True

    def head(
        self,
        url: str
    ) -> None:
        """Make a HEAD request to a single url.

        Parameters
        ----------
        url : str
            The url to which a HEAD request should be made.

        """
        self._url = url
        with self._sess as sess:
            res = sess.head(url)
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["HEAD"] = True

    def options(
        self,
        url: str
    ) -> None:
        """Make an OPTIONS request to a single url.

        Parameters
        ----------
        url : str
            The url to which an OPTIONS request should be made.

        """
        self._url = url
        with self._sess as sess:
            res = sess.options(url)
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["OPTIONS"] = True

    def post(
        self,
        url: str,
        payload: dict = {},
    ) -> None:
        """Make a POST request to a single url.

        Parameters
        ----------
        url : str
            The url to which a POST request should be made.
        payload : dict, optional
            The data which should be delivered with the post request,
            by default {}.

        """
        self._url = url
        with self._sess as sess:
            res = sess.post(
                url,
                data=payload
            )
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["POST"] = True

    def patch(
        self,
        url: str,
    ) -> None:
        """Make a PATCH request to a single url.

        Parameters
        ----------
        url : str
            The url to which a PATCH request should be made.

        """
        self._url = url
        with self._sess as sess:
            res = sess.patch(url)
        # set the attribute
        setattr(self, "_res", res)
        self._http_request["PATCH"] = True

    def parse(
        self,
        name: Optional[str] = None,
        **kwargs: dict,
    ) -> None:
        """Parse a single or a list of response objects.

        Parameters
        ----------
        name : Optional[str]
            Parse only a part of the document specified by a name
            (see the documentation for `_parse_response()`),
            by default None.

        Other Parameters
        ----------------
        Parameters passed into the `SoupStrainer` object, the same as for
        the find_all method of BS4, see the documentation for 
        `_parse_response()`.

        Raises
        ------
        AssertionError
            If `self.load` has not been called before calling this method.

        Notes
        -----
        The parsed response objects are stored in the `data` attribute of the
        class which has type Union[BeautifulSoup, List[BeautifulSoup]]
        depending on the input the same output is returned with parsed htmls.

        """
        if not self._http_request["GET"]:
            raise AssertionError(
                f"Expected {self.get} to be called before calling {self.parse}.")
        if isinstance(self._res, list):
            obj = []
            for response in self._res:
                obj.append(self._parse_response(response, name, **kwargs))
            # with ProcessPoolExecutor(max_workers=self._max_processes) as executor:
            #     obj = list(executor.map(self._parse_response, res))
        else:
            obj = self._parse_response(self._res, name, **kwargs)
        setattr(self, "_data", obj)

    def _parse_response(
        self,
        res: requests.Response,
        name: Optional[str],
        **kwargs: dict,
    ) -> BeautifulSoup:
        """Parse a single response object.

        Parameters
        ----------
        res : requests.Response
            The response to be parsed.
        name : Optional[str]
            Parse only a part of the document specified by a `name`
            (see [1]), by default `None`.

        Other Parameters
        ----------------
        Parameters passed into the `SoupStrainer` object, the same as for
        the `find_all` method of BS4, see [2].

        Returns
        -------
        BeautifulSoup
            The response as Beautifulsoup object.

        Notes
        -----
        This function takes in a string as name which is wrapped
        into a `SoupStrainer` object.

        References
        ----------
        [1] https://www.crummy.com/software/BeautifulSoup/bs4/doc/#soupstrainer
        [2] https://www.crummy.com/software/BeautifulSoup/bs4/doc/#find-all

        """
        # deterime the correct encoding
        http_encoding = res.encoding if "charset" in res.headers.get(
            "content-type", "").lower() else None
        html_encoding = EncodingDetector.find_declared_encoding(
            res.content, is_html=True)
        encoding = html_encoding or http_encoding
        # parse the document
        parse_only = SoupStrainer(name, **kwargs)
        obj = BeautifulSoup(
            res.content,
            self._parser,
            from_encoding=encoding,
            parse_only=parse_only,
        )
        return obj



# -*- coding: utf-8 -*-

import os

import setuptools


def read(fname: str):
    with open(os.path.join(os.path.dirname(__file__), fname), mode="r", encoding="utf-8") as file:
        if fname == "README.md":
            data = file.read()
        else:
            data = file.read().splitlines()
    return data


setuptools.setup(
    name="xscrapers",
    version="0.0.6",
    author="Julian Dwain Stang",
    author_email="julian.stang@web.de",
    description=("A simple webscraping framework."),
    license="MIT",
    url="https://github.com/juliandwain/webscrapers",
    packages=setuptools.find_packages(),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    install_requires=read("requirements.txt"),
    python_requires=">=3.8",
)



# -*- coding: utf-8 -*-

__doc__ = """
"""

from typing import Callable, Dict, List

import pandas as pd
from bs4 import BeautifulSoup, Tag

from ..webscraper import DATA_OBJECT, Webscraper

__all__ = [
    "Parser",
]


def _get_tables(
    tables: list,
) -> List[pd.DataFrame]:
    """Get <table></table> elements as dataframe.

    Parameters
    ----------
    tables : list
        The list of html tables.

    Returns
    -------
    List[pd.DataFrame]
        A list of pandas DataFrames containing the tables.

    """
    df = []
    for table in tables:
        _df = pd.read_html(
            table.prettify(),
            flavor="bs4",
        )[0]
        df.append(_df)
    return df


def _get_href(
    a: list,
    url: str
) -> List[str]:
    """Get href from <a></a> elements as list.

    Parameters
    ----------
    a : list
        The list of html a objects.
    url : str
        The url which is loaded.

    Returns
    -------
    List[str]
        A list of strings containing the urls on the webpage.

    Notes
    -----
    If the href element is not a valid url, then the href element is
    appended to the given url.

    """
    href = []
    for _a in a:
        # if no href attribute is given in the tag, the initial url is returned
        _href = _a.get("href", url)
        if "https" in _href:
            href.append(_href)
        else:
            href.append(url + _href)
    return href


class Parser(Webscraper):

    def __init__(
        self,
        parser: str,
        verbose: bool = False
    ) -> None:
        super().__init__(parser, verbose=verbose)

    def _scrape(
        self,
        tag: str,
        element: DATA_OBJECT,
        fun: Callable,
        *args,
    ) -> Dict[str, list]:
        """Get all html elements defined by `tag`.

        Parameters
        ----------
        tag : str
            The tag which should be searched in the html page.
        element : DATA_OBJECT
            The element to be parsed, this can be:

            * None: If None, then the self._data attribute is parsed.
            * Beautifulsoup: Parse the given Beautifulsoup element.
            * List[Beautifulsoup]: Parse the given Beautifulsoup elements.

        fun : Callable
            The function which loads the element to a specific data type.
        args : tuple
            Additional arguments passed to ``fun``.

        Returns
        -------
        Dict[str, list]
            Return a dictionary containing the url as key and the
            corresponding elements as list.

        Notes
        -----
        If no `element` is given to be searched, then the url(s) is(are)
        searched for only ``tag`` elements.

        Raises
        ------
        AssertionError
            If element is not of type list or Beautifulsoup.

        """
        data = {}
        if not element:
            self.parse(name=tag)
            element = self._data
        if isinstance(element, list):
            if len(element) == 1:
                html_ele = element[0](tag)
                data[self._url] = fun(html_ele, *args)
            else:
                for idx, ele in enumerate(element):
                    # catch if there is a bs4 Tag or ResultSet returned
                    html_ele = ele(tag) if isinstance(
                        ele, Tag) else ele[0](tag)
                    data[self._url[idx]] = fun(html_ele, *args)
        elif isinstance(element, BeautifulSoup):
            html_ele = element(tag)
            data[self._url] = fun(html_ele, *args)
        else:
            raise AssertionError(
                f"Parameter element is not of type {list} nor of type {BeautifulSoup}, it is of type {type(element)}!")
        return data

    def href(
        self,
        element: DATA_OBJECT
    ) -> Dict[str, List[str]]:
        """Get all href of <a></a> elements of the given url(s).

        See the documentation for ``self._scrape()`` for a documentation oif
        additional parameters.
        """
        tag = "a"
        data = self._scrape(tag, element, _get_href, self._url)
        return data

    def table(
        self,
        element: DATA_OBJECT,
    ) -> Dict[str, List[pd.DataFrame]]:
        """Get all <table></table> elements of the given url(s) as
        DataFrame(s).

        See the documentation for ``self._scrape()`` for a documentation oif
        additional parameters.

        """
        tag = "table"
        data = self._scrape(tag, element, _get_tables)
        return data



# -*- coding: utf-8 -*-

import abc
from typing import List, Union
from bs4 import BeautifulSoup

__doc__ = """The base class.
"""

DATA_OBJECT = Union[None, BeautifulSoup, List[BeautifulSoup]]


class Scraper(abc.ABC):
    """The base scraper class.
    """

    def __init__(self, *args, **kwargs):
        self._url = None

    @property
    def url(self) -> Union[None, str, List[str]]:
        """The url object.

        Returns
        -------
        Union[None, str, List[str]]
            The url object.

        """
        return self._url

    @url.setter
    def url(self, val: Union[str, List[str]]) -> None:
        """Set a new url or a list of urls.

        Parameters
        ----------
        val : Union[str, List[str]]
            A single url or a list of urls.

        """
        self._url = val

    @abc.abstractmethod
    def get(self, url: Union[str, List[str]]):
        """Abstract method for getting url content.
        """

    @abc.abstractmethod
    def parse(self):
        """Abstract method for parsing url content.
        """
